/* zalesak.cpp ********************************\
*  FCT scheme with Zalesak's 2/4nd order flux  *
*  written by D.Tselitschev, A.Cherousov       *
*  E-mail: <andrei99@iname.com>                *
*  http://www.geocities.com/andrei_chernousov  *
*  Oct 05, 2003                                *
***********************************************/

#include <cstdio>
#include <cstdlib>
#include <cmath>

#include <iostream>

const double c = 0.5;
  
// 
enum {LOW = 1, HIGH2 = 2, HIGH4 = 4};

enum {LINEAR, NONLINEAR};
int funct;

// dr/dt + df(x)/dx = 0, 
//    f(x) = (c * u) or (c * u * u):
inline double fun(double u)
{
    if (funct)
	return c * u * u; // for nonlinear function it (FCT) sucks!!!
    else 
	return c * u;
}

// maximal of a, b, c
inline double max(double a, double b, double c)
{
  if ((a > b) && (a > c))
    return a;

  if (b > c)
    return b;

  return c;    
} // end max(a, b, c)

// minimal of a, b, c
inline double min(double a, double b, double c)
{
  if ((a < b) && (a < c))
    return a;

  if (b < c)
    return b;

  return c;
} // end min(a, b, c)

// maximal of a, b
inline double max(double a, double b)
{
  if (a > b)
    return a;
  else
    return b;
} // end max(a, b)

// minimal of a, b
inline double min(double a, double b)
{
  if (a < b)
    return a;
  else
    return b;
} // end min(a, b)

/************
*  array1D  *
************/
double *array1d(unsigned size)
{
  double *x;

  // C
  /*
    if ((x = (double *)malloc(size * sizeof(double))) == NULL) {
    fprintf(stderr, "zalesak: can't allocate memory\n");
    exit(-1);
    }
  */
    
  // C++
  x = new double [size]; 

  return x;
   
} // end array1d()

/*
* MAIN
*/
int main()
{
  double 
    *r, *fL, *fH, *rTD, *fAD,
    Pp, Pm, Qp, Qm, 
    *ra, *rmax, *rb, *rmin, *Rp, *Rm, 
    *fC, C; 

  int LEN, order;

  double 
    r_min, r_max,
    deltaX, deltaT, deltaT_X;
  unsigned 
    step, numstep;

  FILE *pF;
  char str[80];
  int i;

  // loading data from file "zalesak.ini" */
  /*
    if ((pF = fopen("zalesak.ini", "r")) == NULL) {
    cerr << "zalesak: can't open file \"zalesak.ini\"\n\n";
    exit(-1);
    }
  */
  pF = stdin;
    
  i = 0;
  do {
    if (fgets(str, 80, pF) != NULL) {
      switch (i) {
      case 0:  LEN    = atoi(str); break;
      case 1:  deltaX = atof(str); break;
      case 2:  deltaT = atof(str); break;
      case 3:  numstep= atoi(str); break;
      case 4:  funct  = atoi(str); break;
      case 5:  order  = atoi(str); break;
      default: break;
      } // end switch()
      i++;
    } // end if()
    else {
      cerr << "zalesak: error reading file \"zalesak.ini\"\n";
      exit(-1);
    }	
  } while (i < 6);
    
  if (pF != stdin)
    fclose(pF);

  // allocate memory dynamically 
  r  = array1d(LEN + 4); 
  rTD= array1d(LEN + 4); 
  fL = array1d(LEN + 1); 
  fH = array1d(LEN + 1); 
  fAD= array1d(LEN + 1);
  ra   = array1d(LEN + 4); 
  rmax = array1d(LEN + 4); 
  rb   = array1d(LEN + 4); 
  rmin = array1d(LEN + 4);
  Rp = array1d(LEN + 4);
  Rm = array1d(LEN + 4); 
  fC = array1d(LEN + 1); 

  //  
  r_min = 0.5;
  r_max = 1.0;

  // initial conditions
  for (i = 2; i <= LEN + 1; i++)
    r[i] = (i > 12 && i <= 22) ? r_max: r_min;

  //
  deltaT_X = deltaT / deltaX;
  //
  step = 0;
    
  // time stepping
  while (step < numstep) {

    if (!(++step%10))
      ;
    // printf("%d\n", step);

    // periodical boundary conditions
    r[    1] = r[LEN+1];
    r[LEN+2] = r[    2];

    //--- higher - 2nd order in space - FCT ---
    if (order == HIGH2 || order == HIGH4) {

      // 1. fL
      for (i = 0; i <= LEN; i++) {
	fL[i] = fun(r[i+1]);
      }	

      // 2. fH: higher-order method
      if (order == HIGH2) { 
	for (i = 0; i <= LEN; i++) { 
	  fH[i] = fun( (r[i+1] + r[i+2]) / 2.0 );
	}
      }
      else { // if (order == HIGH4) {
	for (i = 0; i <= LEN; i++) { 
	  fH[i] = fun( -1.*r[i]/12. + 7.*r[i+1]/12. + 7.*r[i+2]/12. - 1.*r[i+3]/12. );
	}
      }
    
      // 3. rTD: Low order solution
      for (i = 2; i <= LEN+1; i++) {
	rTD[i] = r[i] - deltaT_X * (fL[i-1] - fL[i-2]);
      }
    
      // 4. fAD: Antidiffusion flux
      for (i = 0; i <= LEN; i++) {
	fAD[i] = fH[i] - fL[i];
	// Zalesak's improvement
	if (fAD[i] * (rTD[i+2] - rTD[i+1]) < 0.) 
	    if      (fAD[i] * (rTD[i+3] - rTD[i+2]) < 0.)
		fAD[i] = 0.;
	    else if (fAD[i] * (rTD[i+1] - rTD[i  ]) < 0.)
		fAD[i] = 0.;
      } // end for 4.

      // 5*. Linear uncorrected (test) */
      //for (i = 2; i <= LEN+1; i++) {
      //	r[i] = rTD[i] - deltaT_X * (fAD[i-1] - fAD[i-2]);
      //}

      //     C[i]
      for (i = 2; i <= LEN+1; i++) {
	ra[i] = max(r[i], rTD[i]);
	rb[i] = min(r[i], rTD[i]);
      }
      ra[1]     = ra[LEN+1];
      ra[LEN+2] = ra[2];
      rb[1]     = rb[LEN+1];
      rb[LEN+2] = rb[2];
      //
      for (i = 2; i <= LEN+1; i++) {
	//
	rmax[i] = max(ra[i-1], ra[i], ra[i+1]);
	rmin[i] = min(rb[i-1], rb[i], rb[i+1]);
	// sum of antidiffusion fluxes, entering to cell
	Pp = max(0., fAD[i-2]) - min(0, fAD[i-1]);
	Qp = (rmax[i] - rTD[i]) / deltaX;
	Rp[i] = (Pp > 0.) ? min(1., Qp / Pp) : 0.;
	// sum of antidiffusion fluxes, leaving the cell
	Pm = max(0., fAD[i-1]) - min(0, fAD[i-2]);
	Qm = (rTD[i] - rmin[i] ) / deltaX;
	Rm[i] = (Pm > 0.) ? min(1., Qm / Pm) : 0.; 
      }
	    
      //
      for (i = 0; i <= LEN; i++) {
	// evaluation of C[i]
	C = (fAD[i] >= 0) ? min(Rp[i+2], Rm[i+1]) : min(Rp[i+1], Rm[i+2]);
	// (5) antidiffusion fluxes correction
	fC[i] = C * fAD[i];	
      }

      // (6) Using corrected antidiffusion fluxes
      for (i = 2; i <= LEN+1; i++) {
	r[i] = rTD[i] - deltaT_X * (fC[i-1] - fC[i-2]);
      }	    

    } // end if (order == HIGH2 || ...

    //--- lower - 1nd order donor cell method --- 
    else { // if (order == LOW) {

      // method of donor cell
      for (i = 0; i <= LEN; i++) {
	fL[i] = fun(r[i+1]);
      }

      // 1st order updating
      for (i = 2; i <= LEN+1; i++){
	r[i] = r[i] - deltaT_X * (fL[i-1] - fL[i-2]);
      }

    } // end else 

  } // end while()


  //
  /*
    if ((pF = fopen("input.dat", "w")) == NULL) {
    cerr << "Cannot open file \"input.dat\"\n\n";
    exit(-1);
    }
  */
  pF = stdout;
    
  //fprintf( pF, "%d\n%d\n", LEN+2, 1);
  fprintf( pF, "%1.6f %1.6f \n", 0., r_min-0.1*(r_max-r_min) );
  for (i = 2; i <= LEN+1; i++) {
    fprintf(pF, "%1.6f %1.6f \n", (i-1.5) * deltaX, r[i]);
  }
  fprintf( pF, "%1.6f %1.6f \n", LEN*deltaX, r_max+0.1*(r_max-r_min) );

  if (pF != stdout)
    fclose(pF);

  return 0;
    
} // end main()

